# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mars/serialize/protos/anyref.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='mars/serialize/protos/anyref.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\"mars/serialize/protos/anyref.proto\".\n\x0c\x41nyReference\x12\x0f\n\x07type_id\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x0c\x62\x06proto3')
)




_ANYREFERENCE = _descriptor.Descriptor(
  name='AnyReference',
  full_name='AnyReference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type_id', full_name='AnyReference.type_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='AnyReference.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38,
  serialized_end=84,
)

DESCRIPTOR.message_types_by_name['AnyReference'] = _ANYREFERENCE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AnyReference = _reflection.GeneratedProtocolMessageType('AnyReference', (_message.Message,), dict(
  DESCRIPTOR = _ANYREFERENCE,
  __module__ = 'mars.serialize.protos.anyref_pb2'
  # @@protoc_insertion_point(class_scope:AnyReference)
  ))
_sym_db.RegisterMessage(AnyReference)


# @@protoc_insertion_point(module_scope)
